@@ -25,25 +25,22 @@
 
 #include "mhd_mono_clock.h"
 
-#if defined(_WIN32) && ! defined(__CYGWIN__) && defined(HAVE_CLOCK_GETTIME)
+#if defined(_WIN32) && ! defined(__CYGWIN__)
 /* Prefer native clock source over wrappers */
-#undef HAVE_CLOCK_GETTIME
-#endif /* _WIN32 && ! __CYGWIN__ && HAVE_CLOCK_GETTIME */
-
 #ifdef HAVE_CLOCK_GETTIME
-#include <time.h>
+#undef HAVE_CLOCK_GETTIME
 #endif /* HAVE_CLOCK_GETTIME */
+#ifdef HAVE_GETTIMEOFDAY
+#undef HAVE_GETTIMEOFDAY
+#endif /* HAVE_GETTIMEOFDAY */
+#endif /* _WIN32 && ! __CYGWIN__ */
 
-#ifdef HAVE_GETHRTIME
-#ifdef HAVE_SYS_TIME_H
-/* Solaris defines gethrtime() in sys/time.h */
-#include <sys/time.h>
-#endif /* HAVE_SYS_TIME_H */
 #ifdef HAVE_TIME_H
-/* HP-UX defines gethrtime() in time.h */
 #include <time.h>
 #endif /* HAVE_TIME_H */
-#endif /* HAVE_GETHRTIME */
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif /* HAVE_SYS_TIME_H */
 
 #ifdef HAVE_CLOCK_GET_TIME
 #include <mach/mach.h>
@@ -65,6 +62,10 @@ static clock_serv_t mono_clock_service = _MHD_INVALID_CLOCK_SERV;
 #include <stdint.h>
 #endif /* _WIN32 */
 
+#ifndef NULL
+#define NULL ((void*)0)
+#endif /* ! NULL */
+
 #ifdef HAVE_CLOCK_GETTIME
 #ifdef CLOCK_REALTIME
 #define _MHD_UNWANTED_CLOCK CLOCK_REALTIME
@@ -80,6 +81,10 @@ static clockid_t mono_clock_id = _MHD_UNWANTED_CLOCK;
   defined(HAVE_GETHRTIME)
 static time_t mono_clock_start;
 #endif /* HAVE_CLOCK_GETTIME || HAVE_CLOCK_GET_TIME || HAVE_GETHRTIME */
+#if defined(HAVE_TIMESPEC_GET) || defined(HAVE_GETTIMEOFDAY)
+/* The start value shared for timespec_get() and gettimeofday () */
+static time_t gettime_start;
+#endif /* HAVE_TIMESPEC_GET || HAVE_GETTIMEOFDAY */
 static time_t sys_clock_start;
 #ifdef HAVE_GETHRTIME
 static hrtime_t hrtime_start;
@@ -88,8 +93,8 @@ static hrtime_t hrtime_start;
 #if _WIN32_WINNT >= 0x0600
 static uint64_t tick_start;
 #else  /* _WIN32_WINNT < 0x0600 */
-static int64_t perf_freq;
-static int64_t perf_start;
+static uint64_t perf_freq;
+static uint64_t perf_start;
 #endif /* _WIN32_WINNT < 0x0600 */
 #endif /* _WIN32 */
 
@@ -132,7 +137,7 @@ enum _MHD_mono_clock_source
 
 
 /**
- * Initialise monotonic seconds counter.
+ * Initialise monotonic seconds and milliseconds counters.
  */
 void
 MHD_monotonic_sec_counter_init (void)
@@ -154,6 +159,7 @@ MHD_monotonic_sec_counter_init (void)
      various following ifdef's to work out nicely */
   if (0)
   {
+    (void) 0; /* Mute possible compiler warning */
   }
   else
 #ifdef HAVE_CLOCK_GETTIME
@@ -182,9 +188,23 @@ MHD_monotonic_sec_counter_init (void)
   }
   else
 #endif /* CLOCK_MONOTONIC_COARSE */
+#ifdef CLOCK_MONOTONIC_RAW_APPROX
+  /* Darwin-specific clock */
+  /* Not affected by frequency adjustment, returns clock value cached at
+   * context switch. Can be "milliseconds old", but it's fast. */
+  if (0 == clock_gettime (CLOCK_MONOTONIC_RAW_APPROX,
+                          &ts))
+  {
+    mono_clock_id = CLOCK_MONOTONIC_RAW_APPROX;
+    mono_clock_start = ts.tv_sec;
+    mono_clock_source = _MHD_CLOCK_GETTIME;
+  }
+  else
+#endif /* CLOCK_MONOTONIC_RAW */
 #ifdef CLOCK_MONOTONIC_RAW
-  /* Linux-specific clock */
-  /* Not affected by frequency adjustment, but don't count time in suspend */
+  /* Linux and Darwin clock */
+  /* Not affected by frequency adjustment,
+   * on Linux don't count time in suspend */
   if (0 == clock_gettime (CLOCK_MONOTONIC_RAW,
                           &ts))
   {
@@ -220,6 +240,18 @@ MHD_monotonic_sec_counter_init (void)
   }
   else
 #endif /* CLOCK_BOOTTIME */
+#ifdef CLOCK_UPTIME
+  /* non-Linux clock */
+  /* Doesn't count time in suspend */
+  if (0 == clock_gettime (CLOCK_UPTIME,
+                          &ts))
+  {
+    mono_clock_id = CLOCK_UPTIME;
+    mono_clock_start = ts.tv_sec;
+    mono_clock_source = _MHD_CLOCK_GETTIME;
+  }
+  else
+#endif /* CLOCK_BOOTTIME */
 #endif /* HAVE_CLOCK_GETTIME */
 #ifdef HAVE_CLOCK_GET_TIME
   /* Darwin-specific monotonic clock */
@@ -254,10 +286,10 @@ MHD_monotonic_sec_counter_init (void)
     LARGE_INTEGER freq;
     LARGE_INTEGER perf_counter;
 
-    QueryPerformanceFrequency (&freq);   /* never fail on XP and later */
-    QueryPerformanceCounter (&perf_counter);   /* never fail on XP and later */
-    perf_freq = freq.QuadPart;
-    perf_start = perf_counter.QuadPart;
+    QueryPerformanceFrequency (&freq);       /* never fail on XP and later */
+    QueryPerformanceCounter (&perf_counter); /* never fail on XP and later */
+    perf_freq = (uint64_t) freq.QuadPart;
+    perf_start = (uint64_t) perf_counter.QuadPart;
     mono_clock_source = _MHD_CLOCK_PERFCOUNTER;
   }
   else
@@ -305,12 +337,32 @@ MHD_monotonic_sec_counter_init (void)
   (void) mono_clock_source; /* avoid compiler warning */
 #endif /* HAVE_CLOCK_GET_TIME */
 
+#ifdef HAVE_TIMESPEC_GET
+  if (1)
+  {
+    struct timespec tsg;
+    if (TIME_UTC == timespec_get (&tsg, TIME_UTC))
+      gettime_start = tsg.tv_sec;
+    else
+      gettime_start = 0;
+  }
+#elif defined (HAVE_GETTIMEOFDAY)
+  if (1)
+  {
+    struct timeval tv;
+    if (0 == gettimeofday (&tv, NULL))
+      gettime_start = tv.tv_sec;
+    else
+      gettime_start = 0;
+  }
+#endif /* HAVE_GETTIMEOFDAY */
   sys_clock_start = time (NULL);
 }
 
 
 /**
- * Deinitialise monotonic seconds counter by freeing any allocated resources
+ * Deinitialise monotonic seconds  and milliseconds counters by freeing
+ * any allocated resources
  */
 void
 MHD_monotonic_sec_counter_finish (void)
@@ -327,7 +379,7 @@ MHD_monotonic_sec_counter_finish (void)
 
 
 /**
- * Monotonic seconds counter, useful for timeout calculation.
+ * Monotonic seconds counter.
  * Tries to be not affected by manually setting the system real time
  * clock or adjustments by NTP synchronization.
  *
@@ -364,7 +416,7 @@ MHD_monotonic_sec_counter (void)
     LARGE_INTEGER perf_counter;
 
     QueryPerformanceCounter (&perf_counter);   /* never fail on XP and later */
-    return (time_t) (((uint64_t) (perf_counter.QuadPart - perf_start))
+    return (time_t) (((uint64_t) perf_counter.QuadPart - perf_start)
                      / perf_freq);
   }
 #endif /* _WIN32_WINNT < 0x0600 */
@@ -376,3 +428,77 @@ MHD_monotonic_sec_counter (void)
 
   return time (NULL) - sys_clock_start;
 }
+
+
+/**
+ * Monotonic milliseconds counter, useful for timeout calculation.
+ * Tries to be not affected by manually setting the system real time
+ * clock or adjustments by NTP synchronization.
+ *
+ * @return number of microseconds from some fixed moment
+ */
+uint64_t
+MHD_monotonic_msec_counter (void)
+{
+#if defined(HAVE_CLOCK_GETTIME) || defined(HAVE_TIMESPEC_GET)
+  struct timespec ts;
+#endif /* HAVE_CLOCK_GETTIME || HAVE_TIMESPEC_GET */
+
+#ifdef HAVE_CLOCK_GETTIME
+  if ( (_MHD_UNWANTED_CLOCK != mono_clock_id) &&
+       (0 == clock_gettime (mono_clock_id,
+                            &ts)) )
+    return (uint64_t) (((uint64_t) (ts.tv_sec - mono_clock_start)) * 1000
+                       + (ts.tv_nsec / 1000000));
+#endif /* HAVE_CLOCK_GETTIME */
+#ifdef HAVE_CLOCK_GET_TIME
+  if (_MHD_INVALID_CLOCK_SERV != mono_clock_service)
+  {
+    mach_timespec_t cur_time;
+
+    if (KERN_SUCCESS == clock_get_time (mono_clock_service,
+                                        &cur_time))
+      return (uint64_t) (((uint64_t) (cur_time.tv_sec - mono_clock_start))
+                         * 1000 + (cur_time.tv_nsec / 1000000));
+  }
+#endif /* HAVE_CLOCK_GET_TIME */
+#if defined(_WIN32)
+#if _WIN32_WINNT >= 0x0600
+  if (1)
+    return (uint64_t) (GetTickCount64 () - tick_start);
+#else  /* _WIN32_WINNT < 0x0600 */
+  if (0 != perf_freq)
+  {
+    LARGE_INTEGER perf_counter;
+    uint64_t num_ticks;
+
+    QueryPerformanceCounter (&perf_counter);   /* never fail on XP and later */
+    num_ticks = (uint64_t) (perf_counter.QuadPart - perf_start);
+    return ((num_ticks / perf_freq) * 1000)
+           + ((num_ticks % perf_freq) / (perf_freq / 1000));
+  }
+#endif /* _WIN32_WINNT < 0x0600 */
+#endif /* _WIN32 */
+#ifdef HAVE_GETHRTIME
+  if (1)
+    return ((uint64_t) (gethrtime () - hrtime_start)) / 1000000;
+#endif /* HAVE_GETHRTIME */
+
+  /* Fallbacks, affected by system time change */
+#ifdef HAVE_TIMESPEC_GET
+  if (TIME_UTC == timespec_get (&ts, TIME_UTC))
+    return (uint64_t) (((uint64_t) (ts.tv_sec - gettime_start)) * 1000
+                       + (ts.tv_nsec / 1000000));
+#elif defined (HAVE_GETTIMEOFDAY)
+  if (1)
+  {
+    struct timeval tv;
+    if (0 == gettimeofday (&tv, NULL))
+      return (uint64_t) (((uint64_t) (tv.tv_sec - gettime_start)) * 1000
+                         + (tv.tv_usec / 1000));
+  }
+#endif /* HAVE_GETTIMEOFDAY */
+
+  /* The last resort fallback with very low resolution */
+  return (uint64_t) (time (NULL) - sys_clock_start) * 1000;
+}
