@@ -1,6 +1,7 @@
 /*
  This file is part of libmicrohttpd
  Copyright (C) 2007, 2016 Christian Grothoff
+ Copyright (C) 2016-2021 Evgeny Grin (Karlson2k)
 
  libmicrohttpd is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published
@@ -22,6 +23,7 @@
  * @file mhds_session_info_test.c
  * @brief  Testcase for libmicrohttpd HTTPS connection querying operations
  * @author Sagie Amir
+ * @author Karlson2k (Evgeny Grin)
  */
 
 #include "platform.h"
@@ -55,7 +57,7 @@ query_session_ahc (void *cls, struct MHD_Connection *connection,
 
   if (NULL == *ptr)
   {
-    *ptr = (void*) &query_session_ahc;
+    *ptr = (void *) &query_session_ahc;
     return MHD_YES;
   }
 
@@ -143,7 +145,7 @@ test_query_session ()
     port = (int) dinfo->port;
   }
 
-  if (curl_uses_nss_ssl () == 0)
+  if (curl_tls_is_nss ())
   {
     aes256_sha = "rsa_aes_256_sha";
   }
@@ -217,7 +219,7 @@ main (int argc, char *const *argv)
     curl_global_cleanup ();
     return 77;
   }
-  if (0 != strncmp (ssl_version, "GnuTLS", 6))
+  if (! curl_tls_is_gnutls ())
   {
     fprintf (stderr, "This test can be run only with libcurl-gnutls.\n");
     curl_global_cleanup ();
@@ -228,6 +230,8 @@ main (int argc, char *const *argv)
   curl_global_cleanup ();
   return errorCount != 0 ? 1 : 0;
 #else  /* LIBCURL_VERSION_NUM < 0x072200 */
+  (void) argc; (void) argv;   /* Unused. Silent compiler warning. */
+  (void) query_session_ahc; /* Mute compiler warning */
   return 77;
 #endif /* LIBCURL_VERSION_NUM < 0x072200 */
 }
